from typing import Iterator, Optional, Sized
import math
import copy
import numpy as np

from mmengine.registry import DATA_SAMPLERS
from mmengine.dist import get_dist_info, sync_random_seed
from mmengine.dataset.sampler import InfiniteSampler

from visaionlibrary.datasets.utils import SampleStatus


@DATA_SAMPLERS.register_module()
class VisaionInfiniteSampler(InfiniteSampler):
    def __init__(self,
                 dataset: Sized,
                 shuffle: bool = True,
                 seed: Optional[int] = None,
                 bg_ratio: float = 0.1,
                 batch_size: int = 8
                 ) -> None:
        rank, world_size = get_dist_info()  # get rank and world_size
        self.rank = rank
        self.world_size = world_size

        self.dataset = dataset
        self.shuffle = shuffle
        # make sure the seed is the same in all processes, so that the order of the samples is the same
        self.seed = sync_random_seed() if seed is None else seed
        
        self.bg_ratio = bg_ratio
        self.batch_size = batch_size
        self.fg_sample_indexes, self.bg_sample_indexes = [], []     # store the indexes of foreground and background samples
        for i, sample_type in enumerate(self.dataset.samples_type):    # noqa, samples_type is a list
            if sample_type == SampleStatus.BACKGROUND:
                self.bg_sample_indexes.append(i)
            elif sample_type == SampleStatus.FOREGROUND:
                self.fg_sample_indexes.append(i)
            else:
                raise ValueError(f"sample_type must be one of {SampleStatus.values()}, but got {sample_type}")
        assert len(self.fg_sample_indexes) > 0, "there must be at least one foreground sample in the dataset"

            
        # obtain the number of foreground and background samples in each batch
        assert 0 <= self.bg_ratio < 1, "bg_ratio must be between 0 and 1"
        self.num_bg_per_batch = math.ceil(self.batch_size * self.bg_ratio) if len(self.bg_sample_indexes) > 0 else 0  # once self.bg_ratio is greater than 0, there must be at least one background sample in each batch
        self.num_fg_per_batch = self.batch_size - self.num_bg_per_batch
        assert self.num_fg_per_batch > 0, "there must be at least one foreground sample in each batch"

        self.size = len(self.dataset)  # the number of samples in the dataset, including foreground and background samples

        self.indices = self._indices_of_rank()

    def _infinite_indices(self) -> Iterator[int]:
        """Infinitely yield a sequence of indices.

        """
        np_random = np.random.RandomState(self.seed)  # make sure the random seed is the same in all processes

        # if the number of self.fg_sample_indexes is less than self.batch_size,
        # we need to repeat the self.fg_sample_indexes to fit at least one batch_size
        fg_sample_indexes = copy.copy(self.fg_sample_indexes)
        if len(fg_sample_indexes) < self.batch_size:
            fg_sample_indexes = fg_sample_indexes * math.ceil(self.batch_size / len(fg_sample_indexes))
            fg_sample_indexes = fg_sample_indexes[:self.batch_size]

        # fill the both fg_sample_indexes and bg_sample_indexes to fit one epoch
        num_batches = math.ceil(len(fg_sample_indexes) / self.num_fg_per_batch)  # the number of batches is computed according to the foreground samples
        # make sure the number of batches can be divided by self.world_size
        if num_batches % self.world_size != 0:
            num_batches = num_batches + self.world_size - num_batches % self.world_size

        total_num_fg_samples = num_batches * self.num_fg_per_batch  # the total number of foreground samples needed
        total_num_bg_samples = num_batches * self.num_bg_per_batch  # the total number of background samples needed
        assert total_num_fg_samples >= len(fg_sample_indexes), "Internal error: total_num_fg_samples should always be >= len(self.fg_sample_indexes)"

        # fill the fg_sample_indexes to fit one epoch
        fg_sample_indexes = fg_sample_indexes + \
                            fg_sample_indexes * math.floor((total_num_fg_samples - len(fg_sample_indexes))/len(fg_sample_indexes)) + \
                            np_random.choice(fg_sample_indexes, size=(total_num_fg_samples - len(fg_sample_indexes))%len(fg_sample_indexes)).tolist()

        # fill the bg_sample_indexes to fit one epoch
        if len(self.bg_sample_indexes) > 0 and self.num_bg_per_batch > 0:
            if total_num_bg_samples < len(self.bg_sample_indexes):
                bg_sample_indexes = np_random.choice(self.bg_sample_indexes, size=total_num_bg_samples).tolist()
            else:
                bg_sample_indexes = copy.copy(self.bg_sample_indexes)
                bg_sample_indexes = bg_sample_indexes + \
                                    bg_sample_indexes * math.floor((total_num_bg_samples - len(bg_sample_indexes))/len(bg_sample_indexes)) + \
                                    np_random.choice(bg_sample_indexes, size=(total_num_bg_samples - len(bg_sample_indexes))%len(bg_sample_indexes)).tolist()
        else:
            bg_sample_indexes = []

        while True:
            if self.shuffle:
                np_random.shuffle(fg_sample_indexes)
                np_random.shuffle(bg_sample_indexes)
            
            # rearrange the fg_sample_indexes and bg_sample_indexes according to batch_size and world_size,
            # so that the batch for each process has the same number of foreground and background samples
            # the number of batches for each process is num_batches // world_size
            all_indexes = []
            num_batches_per_process = num_batches // self.world_size
            for i in range(num_batches_per_process):
                all_indexes += fg_sample_indexes[i*(self.num_fg_per_batch*self.world_size):(i+1)*(self.num_fg_per_batch*self.world_size)] + \
                               bg_sample_indexes[i*(self.num_bg_per_batch*self.world_size):(i+1)*(self.num_bg_per_batch*self.world_size)]
            
            yield from all_indexes

@DATA_SAMPLERS.register_module()
class InfiniteSamplerBatch(InfiniteSampler):
    def __init__(self,
                 dataset: Sized,
                 shuffle: bool = True,
                 seed: Optional[int] = None,
                 neg_ratio: float = 0.3,
                 neg_epoch_mode: bool = True,
                 batch_size: int = 8
                 ) -> None:
        rank, world_size = get_dist_info()  # 获取rank和world_size
        self.rank = rank
        self.world_size = world_size

        self.dataset = dataset
        self.shuffle = shuffle
        # make sure the seed is the same in all processes, so that the order of the samples is the same
        self.seed = sync_random_seed() if seed is None else seed

        self.neg_ratio = neg_ratio
        self.batch_size = batch_size
        self.neg_epoch_mode = neg_epoch_mode
        self.pos_data_indexes, self.neg_data_indexes = [], []     # 存储正负样本索引
        for i, sample_type in enumerate(self.dataset.samples_type):    # noqa, samples_type是list, 在det_dataset.py中定义
            if sample_type == SampleStatus.BACKGROUND:
                self.neg_data_indexes.append(i)
            elif sample_type == SampleStatus.FOREGROUND:
                self.pos_data_indexes.append(i)
            else:
                raise ValueError(f"sample_type must be one of {SampleStatus.values()}, but got {sample_type}")

        # get self.pos_num and self.neg_num
        assert 0 <= self.neg_ratio < 1  # 用户可以指定neg_ratio=0但是不能指定成1
        self.neg_num = int(self.batch_size * self.neg_ratio)
        self.pos_num = self.batch_size - self.neg_num

        # 相当于原来的len(self.dataset)
        self.size = math.ceil(len(self.pos_data_indexes) / self.pos_num)    # 向上取整
        self.indices = self._indices_of_rank()

    def _infinite_indices(self) -> Iterator[int]:
        """Infinitely yield a sequence of indices.
        生成器的内容indices
        [
            [pos_index,...,neg_index,..., len(self.pos_num+self.neg_num)],
               ...
        ]
        """
        np_random = np.random.RandomState(self.seed)  # 加这行的原因是numpy.random在多进程环境下重复生成随机数
        neg_data_indexes = copy.copy(self.neg_data_indexes)
        if self.shuffle:
            np_random.shuffle(neg_data_indexes)
        neg_counter = 0
        while True:
            pos_data_indexes = copy.copy(self.pos_data_indexes)
            if self.shuffle:
                np_random.shuffle(pos_data_indexes)
            # 稀有情况下，如果取完全面的几个pos样本正好剩下一个，而中间特征图又出现1x1这种大小的话，就会导致BN层的计算出现问题
            # 所以，需要把pos_data_indexes进行补全，让他至少能够被self._pos_num整除
            # 这个算的是能够让他被整除的最短长度
            target_length = int(np.ceil((len(pos_data_indexes)) / self.pos_num)) * self.pos_num
            # 这一步是把pos_data_indexes拓展到这个目标长度
            pos_data_indexes = (int(np.ceil(target_length / len(pos_data_indexes))) * pos_data_indexes)[:target_length]

            indices: list = []
            for pos_indexes_i in [pos_data_indexes[i:i + self.pos_num]
                                  for i in range(0, len(pos_data_indexes), self.pos_num)]:
                if len(neg_data_indexes) != 0 and (not self.neg_epoch_mode or len(neg_data_indexes) < self.neg_num):
                    neg_indexes_i = np_random.choice(neg_data_indexes, size=self.neg_num).tolist()
                elif len(neg_data_indexes) != 0 and self.neg_epoch_mode:
                    if neg_counter+self.neg_num > len(neg_data_indexes):
                        tem_neg_data_index = neg_data_indexes[neg_counter:]
                        neg_data_indexes = copy.copy(self.neg_data_indexes)
                        if self.shuffle:
                            np_random.shuffle(neg_data_indexes)
                        neg_data_indexes = tem_neg_data_index + neg_data_indexes
                        neg_counter = 0
                    neg_indexes_i = neg_data_indexes[neg_counter: neg_counter+self.neg_num]
                    neg_counter += self.neg_num
                else:
                    neg_indexes_i = []
                indices.append(pos_indexes_i + neg_indexes_i)

            yield from indices



import torch
import torch.nn as nn
import torch.nn.functional as F

from mmcv.cnn import (ConvModule, build_activation_layer, build_conv_layer, build_norm_layer)
from mmcv.cnn.bricks import DropPath
from mmengine.model import BaseModule
from .decode_head import BaseDecodeHead

from mmengine.registry import MODELS

eps = 1e-5


class BasicBlock(BaseModule):
    """
    Similar to BasicBlock in ResNet
    """

    def __init__(self,
                 in_channels,
                 block_channels,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 drop_path_rate=0.0,
                 act_cfg=dict(type='ReLU', inplace=True),
                 init_cfg=None):
        super(BasicBlock, self).__init__(init_cfg=init_cfg)
        self.in_channels = in_channels
        self.block_channels = block_channels
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg

        self.norm1_name, norm1 = build_norm_layer(
            self.norm_cfg, self.block_channels, postfix=1)
        self.norm2_name, norm2 = build_norm_layer(
            self.norm_cfg, self.in_channels, postfix=2)

        self.conv1 = build_conv_layer(
            self.conv_cfg,
            self.in_channels,
            self.block_channels,
            3,
            padding=1,
            bias=False)
        self.add_module(self.norm1_name, norm1)
        self.conv2 = build_conv_layer(
            self.conv_cfg,
            self.block_channels,
            self.in_channels,
            3,
            padding=1,
            bias=False)
        self.add_module(self.norm2_name, norm2)

        self.act = build_activation_layer(self.act_cfg)
        self.drop_path = DropPath(drop_prob=drop_path_rate) if drop_path_rate > eps else nn.Identity()

    @property
    def norm1(self):
        return getattr(self, self.norm1_name)

    @property
    def norm2(self):
        return getattr(self, self.norm2_name)

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.norm1(out)
        out = self.act(out)

        out = self.conv2(out)
        out = self.norm2(out)

        out = self.drop_path(out)

        out += identity

        out = self.act(out)

        return out


class UpSample(BaseModule):

    def __init__(self,
                 in_channels=None,
                 mode='interpolate',
                 interpolate_mode='nearest',
                 upsample_factor=2
                 ):
        super(UpSample, self).__init__()
        assert mode in ['interpolate', 'deconv']
        if mode == 'deconv':
            assert in_channels is not None
        if mode == 'interpolate':
            assert interpolate_mode in ['nearest', 'bilinear', 'bicubic']

        self.in_channels = in_channels
        self.mode = mode
        self.upsample_factor = upsample_factor
        self.interpolate_mode = interpolate_mode

        if mode == 'interpolate':
            self.upsample = nn.Upsample(scale_factor=self.upsample_factor, mode=self.interpolate_mode)
        elif mode == 'deconv':
            self.upsample = nn.ConvTranspose2d(in_channels=self.in_channels,
                                               out_channels=self.in_channels,
                                               kernel_size=3,
                                               stride=self.upsample_factor,
                                               padding=1,
                                               bias=True)
        else:
            raise ValueError

    def forward(self, x):
        if self.mode == 'interpolate':
            out = self.upsample(x)
        elif self.mode == 'deconv':
            shape = list(x.shape)
            shape[-2] = shape[-2] * self.upsample_factor
            shape[-1] = shape[-1] * self.upsample_factor
            out = self.upsample(x, output_size=shape)
        else:
            raise ValueError

        return out


class DecoderStage(BaseModule):
    def __init__(self,
                 in_channels,
                 block_channels,
                 skip_channels=0,
                 num_blocks=1,
                 block_drop_path_rate=0.,
                 upsample_mode='interpolate',
                 upsample_interpolate_mode='nearest',
                 upsample_factor=2,
                 enable_fusion=True,
                 fusion_mode='concate',
                 fusion_density_mode='common',
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU', inplace=True),
                 ):
        super(DecoderStage, self).__init__()
        self.in_channels = in_channels
        self.block_channels = block_channels
        self.skip_channels = skip_channels
        self.num_blocks = num_blocks
        self.block_drop_path_rate = block_drop_path_rate
        self.upsample_mode = upsample_mode
        self.upsample_interpolate_mode = upsample_interpolate_mode
        self.upsample_factor = upsample_factor
        self.enable_fusion = enable_fusion
        assert fusion_mode in ['concate', 'add']
        self.fusion_mode = fusion_mode
        assert fusion_density_mode in ['common', 'dense']
        self.fusion_density_mode = fusion_density_mode
        # 当fusion_mode = concate 的情况下， fusion_density_mode 只能是 common
        if self.fusion_mode == 'concate':
            assert self.fusion_density_mode == 'common'

        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg

        # 如果是add，需要把skip和input的channel数对齐
        if self.enable_fusion:
            if self.fusion_mode == 'add':
                self.skip_align = ConvModule(
                    in_channels=self.skip_channels,
                    out_channels=self.block_channels,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    dilation=1,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=self.act_cfg
                )
                self.input_align = ConvModule(
                    in_channels=self.in_channels,
                    out_channels=self.block_channels,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    dilation=1,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=self.act_cfg
                )
            else:  # self.fusion_mode == 'concate'
                self.concate_align = ConvModule(
                    in_channels=self.in_channels + self.skip_channels,
                    out_channels=self.block_channels,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    dilation=1,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=self.act_cfg
                )
        else:
            self.input_align = ConvModule(
                in_channels=self.in_channels,
                out_channels=self.block_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                dilation=1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg
            )

        self._make_stage()
        assert len(self.stage_list) == (num_blocks + 1)

    def _make_stage(self):
        self.stage_list = nn.ModuleList()
        for i in range(self.num_blocks):
            block = BasicBlock(
                in_channels=self.block_channels,
                block_channels=self.block_channels,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                drop_path_rate=self.block_drop_path_rate,
                act_cfg=self.act_cfg,
            )
            self.stage_list.append(block)

        upsample = UpSample(
            in_channels=self.block_channels,
            mode=self.upsample_mode,
            interpolate_mode=self.upsample_interpolate_mode,
            upsample_factor=self.upsample_factor
        )
        self.stage_list.append(upsample)

    def forward(self, x, skip=None):
        if skip is None:
            assert not self.enable_fusion

        if self.enable_fusion:
            if self.fusion_mode == 'add':
                skip_align = self.skip_align(skip)
                x = self.input_align(x)

            for i in range(self.num_blocks):
                if self.fusion_mode == 'concate':
                    if i == 0:
                        assert self.fusion_density_mode == 'common'
                        x = torch.cat([x, skip], dim=1)
                        x = self.concate_align(x)
                        x = self.stage_list[i](x)
                    else:
                        x = self.stage_list[i](x)
                else:  # self.fusion_mode == 'add'
                    if self.fusion_density_mode == 'common':
                        if i == 0:
                            x = x + skip_align
                        x = self.stage_list[i](x)
                    else:  # self.fusion_density_mode == 'dense'
                        x = x + skip_align
                        x = self.stage_list[i](x)
        else:
            x = self.input_align(x)
            for i in range(self.num_blocks):
                x = self.stage_list[i](x)

        out = self.stage_list[-1](x)
        return out


@MODELS.register_module()
class HYHUNetHead(BaseDecodeHead):
    def __init__(self,
                 prediction_channels=128,
                 num_classes=2,
                 fusion_mode='concate',
                 fusion_density_mode='common',
                 block_drop_path_rate=0.,
                 upsample_mode='interpolate',
                 upsample_interpolate_mode='nearest',
                 num_block_list=[1, 1, 1, 1],
                 block_channel_list=[256, 128, 64, 32],
                 input_channel_list=[64, 128, 256, 512],
                 last_scale_factor=2,
                 is_dual=False,
                 dual_weight=10,
                 threshold=0.5,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU', inplace=True),
                 init_cfg=[  # 怎么设置：https://mmengine.readthedocs.io/en/latest/advanced_tutorials/initialize.html
                     dict(type='Kaiming', layer=['Conv2d', 'ConvTranspose2d']),
                 ],
                 **kwargs
                 ):
        super().__init__(in_channels=0, channels=prediction_channels, num_classes=num_classes, init_cfg=init_cfg, **kwargs)

        assert len(input_channel_list) == len(block_channel_list)
        assert len(block_channel_list) == len(num_block_list)

        self.prediction_channels = prediction_channels
        self.num_classes = num_classes
        self.fusion_mode = fusion_mode
        self.fusion_density_mode = fusion_density_mode
        self.block_drop_path_rate = block_drop_path_rate
        self.upsample_mode = upsample_mode
        self.upsample_interpolate_mode = upsample_interpolate_mode
        self.num_block_list = num_block_list
        self.block_channel_list = block_channel_list
        self.input_channel_list = input_channel_list
        self.last_scale_factor = last_scale_factor
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.is_dual = is_dual
        self.dual_weight = dual_weight
        self.num_stages = len(num_block_list)
        self.decoder_stages = nn.ModuleList()
        for i in range(self.num_stages):
            stage = DecoderStage(
                in_channels=self.input_channel_list[self.num_stages - 1] if i == 0 else self.block_channel_list[i - 1],
                block_channels=self.block_channel_list[i],
                skip_channels=0 if i == 0 else self.input_channel_list[self.num_stages - 1 - i],
                num_blocks=self.num_block_list[i],
                block_drop_path_rate=self.block_drop_path_rate,
                upsample_mode=self.upsample_mode,
                upsample_interpolate_mode=self.upsample_interpolate_mode,
                upsample_factor=self.last_scale_factor if i == (self.num_stages - 1) else 2,
                enable_fusion=False if i == 0 else True,
                fusion_mode=self.fusion_mode,
                fusion_density_mode=self.fusion_density_mode,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg
            )
            self.decoder_stages.append(stage)

        # 构造预测层
        self.final_conv = ConvModule(
            in_channels=self.block_channel_list[-1],
            out_channels=self.prediction_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            dilation=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg
        )
        self.conv_seg = nn.Conv2d(self.prediction_channels, self.out_channels, kernel_size=(1, 1), padding=0)

    def forward(self, feature_maps):
        assert len(feature_maps) == self.num_stages
        last_feature_map = feature_maps[-1]
        for i in range(self.num_stages):
            if i == 0:
                out = self.decoder_stages[i](last_feature_map)
            else:
                out = self.decoder_stages[i](out, feature_maps[self.num_stages - 1 - i])
        out = self.final_conv(out)
        out = self.conv_seg(out)
        return out

    @staticmethod
    def tran_sim_loss(seg_logits, batch_data_samples):
        batch_size = seg_logits.size(0) // 2
        real_seg_logits = seg_logits[:batch_size]
        dual_seg_logits = seg_logits[batch_size:]
        all_kl_loss = []
        for i in range(batch_size):
            if len(batch_data_samples[i].overlap_region) == 0:
                continue
            x_loc_orig, y_loc_orig, x_loc_dual, y_loc_dual = batch_data_samples[i].overlap_region
            real_seg_logits_overlap = real_seg_logits[i, :, y_loc_orig, x_loc_orig]
            dual_seg_logits_overlap = dual_seg_logits[i, :, y_loc_dual, x_loc_dual]
            kl = F.kl_div((dual_seg_logits_overlap.softmax(dim=0) + 1e-10).log(),
                          (real_seg_logits_overlap.softmax(dim=0) + 1e-10), reduction='mean')
            all_kl_loss.append(kl)
        if len(all_kl_loss) > 0:
            mean_kl_loss = sum(all_kl_loss) / len(all_kl_loss)
            return mean_kl_loss
        else:
            return 0

    def loss(self, inputs, batch_data_samples, train_cfg):
        seg_logits = self.forward(inputs)
        losses = self.loss_by_feat(seg_logits, batch_data_samples)
        if self.is_dual:
            losses['loss_match'] = self.tran_sim_loss(seg_logits, batch_data_samples) * self.dual_weight
        return losses



import os
import numpy as np
from PIL import Image
import torch

from edge_sam import sam_model_registry, SamPredictor

class EdgeSAM:
    def __init__(self, model_path="edge_sam.pth", device="cuda:0"):
        VISAION_DIR = os.getenv("VISAION_DIR")
        if VISAION_DIR is None:
            raise ValueError("VISAION_DIR is not set")
        self.model_path = os.path.join(VISAION_DIR, "weights", model_path)
        self.device = device

        model_type = "edge_sam"
        self.sam = sam_model_registry[model_type](checkpoint=self.model_path)
        self.sam.to(device=self.device)

        self.predictor = SamPredictor(self.sam)


    def predict_by_points(self, images:list[Image.Image], points:list[np.ndarray]=None, labels:list[int]=None, num_multimask_outputs:int=1):
        """
        Predict the masks of the images.
        Args:
            images: list[Image.Image]
            points: list[np.ndarray]
            labels: list[int]
            num_multimask_outputs: int
        Returns:
            masks_list: list[np.ndarray]
            scores_list: list[np.ndarray]
            logits_list: list[np.ndarray]
        """
        if len(images) == 0:
            raise ValueError("The length of images must be greater than 0")
        
        for image, point_coords, point_labels in zip(images, points, labels):
            assert point_coords is not None and point_labels is not None, "point_coords and point_labels must be provided"
            assert len(point_coords) == len(point_labels), "the length of point_coords and point_labels must be the same"
            assert len(point_coords) > 0, "the length of point_coords must be greater than 0"
            assert len(point_coords) == len(point_labels), "the length of point_coords and point_labels must be the same"

            input_image = np.array(image)   # Image2numpy for EdgeSAM
            self.predictor.set_image(input_image)
            masks, scores, logits = self.predictor.predict(
                point_coords=point_coords,
                point_labels=point_labels,
                num_multimask_outputs=num_multimask_outputs,
                use_stability_score=True
            )

            yield masks, scores, logits


    def predict_by_bboxes(self, images:list[Image.Image], bboxes:list[np.ndarray]=None, num_multimask_outputs:int=1):
        """
        Predict the masks of the images.
        Args:
            images: list[Image.Image]
            bboxes: list[np.ndarray]
            num_multimask_outputs: int
        Returns:
            masks_list: list[np.ndarray]
            scores_list: list[np.ndarray]
            logits_list: list[np.ndarray]
        """
        if len(images) == 0:
            raise ValueError("The length of images must be greater than 0")
        
        for image, input_bboxes in zip(images, bboxes):
            assert input_bboxes is not None, "bbox must be provided"

            input_image = np.array(image)   # Image2numpy for EdgeSAM
            self.predictor.set_image(input_image)
            input_box = torch.tensor(input_bboxes, device=self.device)
            transformed_boxes = self.predictor.transform.apply_boxes_torch(input_box, input_image.shape[:2])
            masks, scores, logits = self.predictor.predict_torch(
                point_coords=None,
                point_labels=None,
                boxes=transformed_boxes,
                num_multimask_outputs=num_multimask_outputs,
                features=self.predictor.features,
            )

            masks = masks.cpu().numpy() if isinstance(masks, torch.Tensor) else masks
            scores = scores.cpu().numpy() if isinstance(scores, torch.Tensor) else scores
            logits = logits.cpu().numpy() if isinstance(logits, torch.Tensor) else logits
            
            yield masks, scores, logits



# Copyright (c) . All rights reserved.

from cycler import K
from mmyolo.models.backbones import YOLOv8CSPDarknet
from mmyolo.models.utils import make_divisible, make_round
from mmengine.registry import MODELS

from ..layers.yolo_bricks import C3K2, C2PSA, Conv, SPPF



@MODELS.register_module()
class YOLOv11CSPDarknet(YOLOv8CSPDarknet):
    """
    在YOLOv8 backbone基础上, 将csp layer换为C3K2, 最后再加上一个C2PSA
    """
    arch_settings = {
        'P5': [[64, 128, 2, 0.25, False, False], [128, 256, 2, 0.25, False, False],
               [256, 512, 2, 0.5, True, False], [512, 1024, 2, 0.5, True, True]],
    } # blocks different with v8

    def build_stem_layer(self):
         return Conv(
            self.input_channels,
            make_divisible(self.arch_setting[0][0], self.widen_factor),
            k=3,
            s=2)

    def build_stage_layer(self, stage_idx: int, setting: list) -> list:
        """Build a stage layer.

        Args:
            stage_idx (int): The index of a stage layer.
            setting (list): The architecture setting of a stage layer.
        """
        in_channels, out_channels, num_blocks, expand_ratio, c3k, use_spp = setting
        if self.widen_factor > 0.5: # foe xlm model size
            c3k = True
        in_channels = make_divisible(in_channels, self.widen_factor)
        out_channels = make_divisible(out_channels, self.widen_factor)
        num_blocks = make_round(num_blocks, self.deepen_factor)
        c3k_in_channels = out_channels
        if stage_idx <2:
            c3k_out_channels= out_channels*2
        else:
            c3k_out_channels= out_channels
        if stage_idx == 1 or stage_idx == 2:
            in_channels = out_channels
        stage = []
        conv_layer = Conv(
            in_channels,
            out_channels,
            k=3,
            s=2)
        stage.append(conv_layer)
        csp_layer = C3K2(
            c3k_in_channels,
            c3k_out_channels,
            num_blocks=num_blocks,
            expand_ratio=expand_ratio,
            c3k=c3k)
        stage.append(csp_layer)
        if use_spp: # last stage
            spp = SPPF(
                out_channels,
                out_channels,
                k=5,)
            stage.append(spp)
            c2psa = C2PSA(out_channels, out_channels, n=num_blocks)  
            stage.append(c2psa)
        return stage


from visaionlibrary.evaluation.metrics import (
    PrecisionAndRecallsMetricforVisaion, 
    PrecisionAndRecallsMetric, 
    AccuracyVisaion, 
    VisaionMetric, 
    VisaionMetricHttp, 
    TransStabilityMetric,
    CocoMetricVisaion,
)

__all__ = [
    "PrecisionAndRecallsMetricforVisaion", 
    "PrecisionAndRecallsMetric", 
    "AccuracyVisaion", 
    "VisaionMetric", 
    "VisaionMetricHttp", 
    "TransStabilityMetric", 
    "CocoMetricVisaion",
]


from visaionlibrary.models.seg import EncoderDecoder, BaseDecodeHead, HYHUNetHead, RegNetVisaion
from visaionlibrary.models.data_preprocessor import SegDataPreProcessor, DetDataPreprocessor
from visaionlibrary.models.ins import VisaionRTMDetInsSepBNHead, YOLO11InsHeadModule
from visaionlibrary.models.backbones.csp_darknet import YOLOv11CSPDarknet
from visaionlibrary.models.necks.yolo11_pafpn import YOLO11PAFPN
from visaionlibrary.models.dense_heads.yolo11_head import YOLO11HeadModule

__all__ = ["EncoderDecoder", "BaseDecodeHead", "HYHUNetHead", "RegNetVisaion", "SegDataPreProcessor", "DetDataPreprocessor", "VisaionRTMDetInsSepBNHead",
    "YOLOv11CSPDarknet", "YOLO11PAFPN", "YOLO11HeadModule", "YOLO11InsHeadModule"]



# This file was generated by Nuitka

# Stubs included by default
from __future__ import annotations
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad
from enum import Enum
from mmdet.apis import init_detector
from mmengine.config import Config
from mmengine.logging import MMLogger, print_log
from mmengine.model.utils import revert_sync_batchnorm
from mmengine.registry import init_default_scope
from mmpretrain.apis import init_model
from mmseg.apis import init_model
from torch import Tensor
from typing import Any, Callable, Dict, List, Optional, Sequence, Union
from typing_extensions import Self
from visaionlibrary import __version__
import copy
import json
import numpy
import onnx
import onnxruntime
import onnxsim
import os
import os.path
import platform
import tempfile
import time
import torch
import torch.nn
import torch.nn.functional

class YoloDetOnnxWrapper:
    def __init__(self: Self, config: Any, checkpoint: Any, device: Any) -> None: ...
    def _predict_by_feat(self: Self, out: Any) -> Any: ...
    def forward(self: Self, inputs: Any) -> Any: ...

class SegOnnxWrapper:
    def __init__(self: Self, config: Any, checkpoint: Any, device: Any) -> None: ...
    def forward(self: Self, inputs: Any) -> Any: ...

class Yolov5InsOnnxWrapper:
    def __init__(self: Self, config: Any, checkpoint: Any, device: Any) -> None: ...
    def _predict_by_feat(self: Self, cls_scores: Sequence[Tensor], bbox_preds: Sequence[Tensor], objectnesses: Sequence[Tensor], coeff_preds: Sequence[Tensor], proto_preds: Tensor) -> Any: ...
    def forward(self: Self, inputs: Any) -> Any: ...

class ResnetOnnxWrapper:
    def __init__(self: Self, config: Any, checkpoint: Any, device: Any) -> None: ...
    def forward(self: Self, inputs: Any) -> Any: ...

class RTMDetInsOnnxWrapper:
    def __init__(self: Self, config: Any, checkpoint: Any, device: Any) -> None: ...
    def _parse_dynamic_params(self: Self, flatten_kernels: Any) -> Any: ...
    def _mask_predict_by_feat_single(self: Self, mask_feat: Any, kernels: Any, priors: Any) -> Any: ...
    def _predict_by_feat(self: Self, cls_scores: List[Tensor], bbox_preds: List[Tensor], kernel_preds: List[Tensor], mask_feat: Tensor, score_factors: Optional[List[Tensor]]) -> Any: ...
    def forward(self: Self, inputs: Any) -> Any: ...

class AESCipher:
    def __init__(self: Self, k: bytes, mode: int) -> None: ...
    def encrypt(self: Self, raw: bytes) -> hex: ...
    def decrypt(self: Self, encrypted_data: hex) -> bytes: ...

def build_logger(work_dir: str, log_level: Union[int, str]) -> MMLogger:
    ...

def _extract_baseParam_visaion_keys(data: Any, result: Any) -> Any:
    ...

def _extract_scale_factor(cfg: list) -> Any:
    ...

def _extract_crop_size_and_stride(cfg: dict) -> Any:
    ...

def _extract_crop_size_from_YOLOv5KeepRatioResize(cfg: dict) -> Any:
    ...

def _extract_crop_size_from_Resize(cfg: dict) -> Any:
    ...

def export_entry(job_info: dict) -> Any:
    ...


__name__ = ...



# Modules used internally, to allow implicit dependencies to be seen:
import typing
import os
import ntpath
import numpy
import time
import copy
import json
import platform
import tempfile
import enum
import Crypto
import Crypto.Cipher
import Crypto.Cipher.AES
import Crypto.Util
import Crypto.Util.Padding
import torch
import torch.Tensor
import torch.nn
import torch.nn.functional
import onnx
import onnxsim
import onnxruntime
import mmseg
import mmseg.apis
import mmseg.apis.init_model
import mmdet
import mmdet.apis
import mmdet.apis.init_detector
import mmpretrain
import mmpretrain.apis
import mmpretrain.apis.init_model
import mmengine
import mmengine.model
import mmengine.model.utils
import mmengine.logging
import mmengine.logging.MMLogger
import mmengine.logging.print_log
import mmengine.config
import mmengine.config.Config
import mmengine.registry
import mmengine.registry.init_default_scope
import visaionlibrary
import visaionlibrary.__version__


import os
from pathlib import Path

def increment_path(path, exist_ok=False, sep="", mkdir=False):
    """
    Increments a file or directory path, i.e., runs/exp --> runs/exp{sep}2, runs/exp{sep}3, ... etc.

    If the path exists and `exist_ok` is not True, the path will be incremented by appending a number and `sep` to
    the end of the path. If the path is a file, the file extension will be preserved. If the path is a directory, the
    number will be appended directly to the end of the path. If `mkdir` is set to True, the path will be created as a
    directory if it does not already exist.

    Args:
        path (str | pathlib.Path): Path to increment.
        exist_ok (bool): If True, the path will not be incremented and returned as-is.
        sep (str): Separator to use between the path and the incrementation number.
        mkdir (bool): Create a directory if it does not exist.

    Returns:
        (pathlib.Path): Incremented path.

    Examples:
        Increment a directory path:
        >>> from pathlib import Path
        >>> path = Path("runs/exp")
        >>> new_path = increment_path(path)
        >>> print(new_path)
        runs/exp2

        Increment a file path:
        >>> path = Path("runs/exp/results.txt")
        >>> new_path = increment_path(path)
        >>> print(new_path)
        runs/exp/results2.txt
    """
    path = Path(path)  # os-agnostic
    if path.exists() and not exist_ok:
        path, suffix = (path.with_suffix(""), path.suffix) if path.is_file() else (path, "")

        # Method 1
        for n in range(2, 9999):
            p = f"{path}{sep}{n}{suffix}"  # increment path
            if not os.path.exists(p):
                break
        path = Path(p)

    if mkdir:
        path.mkdir(parents=True, exist_ok=True)  # make directory

    return path



import math
import time
import torch
import torch.nn as nn
import torch.nn.functional as F

from visaionlibrary.utils.yoloe.utils.checks import check_version


# Version checks (all default to version>=min_version)
TORCH_1_9 = check_version(torch.__version__, "1.9.0")
TORCH_1_13 = check_version(torch.__version__, "1.13.0")



def smart_inference_mode():
    """Applies torch.inference_mode() decorator if torch>=1.9.0 else torch.no_grad() decorator."""

    def decorate(fn):
        """Applies appropriate torch decorator for inference mode based on torch version."""
        if TORCH_1_9 and torch.is_inference_mode_enabled():
            return fn  # already in inference_mode, act as a pass-through
        else:
            return (torch.inference_mode if TORCH_1_9 else torch.no_grad)()(fn)

    return decorate


def time_sync():
    """PyTorch-accurate time."""
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    return time.time()


def fuse_conv_and_bn(conv, bn):
    """Fuse Conv2d() and BatchNorm2d() layers https://tehnokv.com/posts/fusing-batchnorm-and-conv/."""
    fusedconv = (
        nn.Conv2d(
            conv.in_channels,
            conv.out_channels,
            kernel_size=conv.kernel_size,
            stride=conv.stride,
            padding=conv.padding,
            dilation=conv.dilation,
            groups=conv.groups,
            bias=True,
        )
        .requires_grad_(False)
        .to(conv.weight.device)
    )

    # Prepare filters
    w_conv = conv.weight.view(conv.out_channels, -1)
    w_bn = torch.diag(bn.weight.div(torch.sqrt(bn.eps + bn.running_var)))
    fusedconv.weight.copy_(torch.mm(w_bn, w_conv).view(fusedconv.weight.shape))

    # Prepare spatial bias
    b_conv = torch.zeros(conv.weight.shape[0], device=conv.weight.device) if conv.bias is None else conv.bias
    b_bn = bn.bias - bn.weight.mul(bn.running_mean).div(torch.sqrt(bn.running_var + bn.eps))
    fusedconv.bias.copy_(torch.mm(w_bn, b_conv.reshape(-1, 1)).reshape(-1) + b_bn)

    return fusedconv


def fuse_deconv_and_bn(deconv, bn):
    """Fuse ConvTranspose2d() and BatchNorm2d() layers."""
    fuseddconv = (
        nn.ConvTranspose2d(
            deconv.in_channels,
            deconv.out_channels,
            kernel_size=deconv.kernel_size,
            stride=deconv.stride,
            padding=deconv.padding,
            output_padding=deconv.output_padding,
            dilation=deconv.dilation,
            groups=deconv.groups,
            bias=True,
        )
        .requires_grad_(False)
        .to(deconv.weight.device)
    )

    # Prepare filters
    w_deconv = deconv.weight.view(deconv.out_channels, -1)
    w_bn = torch.diag(bn.weight.div(torch.sqrt(bn.eps + bn.running_var)))
    fuseddconv.weight.copy_(torch.mm(w_bn, w_deconv).view(fuseddconv.weight.shape))

    # Prepare spatial bias
    b_conv = torch.zeros(deconv.weight.shape[1], device=deconv.weight.device) if deconv.bias is None else deconv.bias
    b_bn = bn.bias - bn.weight.mul(bn.running_mean).div(torch.sqrt(bn.running_var + bn.eps))
    fuseddconv.bias.copy_(torch.mm(w_bn, b_conv.reshape(-1, 1)).reshape(-1) + b_bn)

    return fuseddconv


def initialize_weights(model):
    """Initialize model weights to random values."""
    for m in model.modules():
        t = type(m)
        if t is nn.Conv2d:
            pass  # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
        elif t is nn.BatchNorm2d:
            m.eps = 1e-3
            m.momentum = 0.03
        elif t in {nn.Hardswish, nn.LeakyReLU, nn.ReLU, nn.ReLU6, nn.SiLU}:
            m.inplace = True


def scale_img(img, ratio=1.0, same_shape=False, gs=32):
    """Scales and pads an image tensor, optionally maintaining aspect ratio and padding to gs multiple."""
    if ratio == 1.0:
        return img
    h, w = img.shape[2:]
    s = (int(h * ratio), int(w * ratio))  # new size
    img = F.interpolate(img, size=s, mode="bilinear", align_corners=False)  # resize
    if not same_shape:  # pad/crop img
        h, w = (math.ceil(x * ratio / gs) * gs for x in (h, w))
    return F.pad(img, [0, w - s[1], 0, h - s[0]], value=0.447)  # value = imagenet mean


def intersect_dicts(da, db, exclude=()):
    """Returns a dictionary of intersecting keys with matching shapes, excluding 'exclude' keys, using da values."""
    return {k: v for k, v in da.items() if k in db and all(x not in k for x in exclude) and v.shape == db[k].shape}



from typing import List
import tempfile
import os
import os.path as osp
from pathlib import Path

from mmengine.registry import DATASETS
from mmengine.logging import print_log
from mmseg.datasets import BaseSegDataset
from mmengine import load

from visaionlibrary.datasets.utils import SampleStatus

@DATASETS.register_module()
class VisaionSegDataset(BaseSegDataset):
    def load_data_list(self) -> List[dict]:
        # self.data_root格式校验
        if isinstance(self.data_root, str) and "::" not in self.data_root:
                self.data_root = [self.data_root]
        elif isinstance(self.data_root, str) and "::" in self.data_root:
            self.data_root = self.data_root.split("::")
        elif isinstance(self.data_root, list):
            pass
        else:
            raise ValueError(f"data_root must be a string of dataset path or a list of dataset path, but got {self.data_root}")
        
        # dataset可以是文件夹和json文件
        # - 文件夹是visaion后端格式
        # - json是visaion前端兼容格式
        if self.data_root[0].endswith("json"):
            data_list = self.load_seg_from_json()
        else:
            data_list = self.load_seg_from_folder()
        
        # 判断数据集中每个数据是前景还是背景
        print_log(f"collect samples type ...", "current")
        self.samples_type = []
        for data in data_list:
            if data['seg_map_path'] is None:
                self.samples_type.append(SampleStatus.BACKGROUND)
            else:
                self.samples_type.append(SampleStatus.FOREGROUND)

        return data_list
    
    def load_seg_from_folder(self) -> List[dict]:
        """
        提取数据集文件夹中的路径信息

        每个数据集情况
            |-- image1.bmp
            |-- image1.json
            |-- *.bmp
            |-- *.json
            |-- label.json
            |-- background.json
        
        Args:
            None
        Returns:
            data_list: List[dict]   返回数据集中每个数据的dict
                [
                    {
                        "dataset_metainfo":{"classes":[], "color":[]},
                        "img_path": image_path,
                        "seg_map_path": mixed_json_path,
                        "sample_global_id": str,
                        "label_map":None,
                        'reduce_zero_label':True,
                        'seg_fields':[]
                    },
                    ...
                ]
        """
        
        # 读取多个数据集
        data_list = []          # 存储最终的结果
        img_suffix = ['.bmp', '.jpg', '.jpeg', '.png']

        for dataset_dir in self.data_root:
            # 读取背景信息
            background_info_path = osp.join(dataset_dir, 'background.json')
            background_info = load(background_info_path) if osp.exists(background_info_path) else []

            # 读取正样本
            images_path = [osp.join(dataset_dir,t) for t in os.listdir(dataset_dir) if Path(t).suffix in img_suffix]
            for img_p in images_path:
                is_background = True if os.path.basename(img_p) in background_info else False

                mask_p = img_p.replace(Path(img_p).suffix, '.json')
                if not osp.exists(mask_p):
                    mask_p = img_p.replace(Path(img_p).suffix, '.mix.json')
                mask_path = mask_p if osp.exists(mask_p) else None

                # one_record存放一条记录
                one_record = dict()
                one_record["dataset_metainfo"] = self._metainfo
                one_record["img_path"] = img_p
                one_record["seg_map_path"] = mask_path
                one_record['label_map'] = self.label_map
                one_record['reduce_zero_label'] = self.reduce_zero_label
                one_record['seg_fields'] = []
                if mask_path is not None:   # 有json文件，那就是已标注样本
                    one_record['sample_type'] = SampleStatus.FOREGROUND
                elif is_background:         # 没有json但是在background里面，那就是背景样本
                    one_record['sample_type'] = SampleStatus.BACKGROUND
                else:                       # 没有json也没有bg信息，那就是无标注样本
                    one_record['sample_type'] = SampleStatus.RAW
                data_list.append(one_record)

        return data_list
    
    def load_seg_from_json(self) -> List[dict]:
        """
        提取数据集json文件中的路径信息
        """
        VISAION_DIR = os.environ['VISAION_DIR']
        assert VISAION_DIR is not None, "VISAION_DIR is not set"

        # 读取多个数据集
        data_list = []          # 存储最终的结果

        for dataset_json_path in self.data_root:
            # 读取数据集内容, 即一个json文件中的内容
            dataset_json_dict = load(dataset_json_path)
            all_sample = dataset_json_dict['dataset']
            url_prefix = dataset_json_dict['meta_info']['url_prefix']

            for one_sample in all_sample:
                # 获取图片和标注路径
                image_path = osp.join(VISAION_DIR, url_prefix, one_sample['data'][0]['img_path'])
                anno_path = one_sample.get('mix_annotation_path', None)
                if anno_path is not None and anno_path != '':
                    anno_path = osp.join(VISAION_DIR, url_prefix, anno_path)
                anno_path = None if not osp.exists(anno_path) else anno_path

                # one_record存放一条记录
                one_record = dict()
                one_record["dataset_metainfo"] = self._metainfo
                one_record["img_path"] = image_path
                one_record["seg_map_path"] = anno_path
                one_record['label_map'] = self.label_map
                one_record['reduce_zero_label'] = self.reduce_zero_label
                one_record['seg_fields'] = []
                if one_sample.get('status', SampleStatus.RAW.value) == SampleStatus.RAW.value:
                    raise ValueError(f"sample {one_sample['data'][0]['img_path']} is raw")
                one_record['sample_type'] = SampleStatus(one_sample.get('status', SampleStatus.BACKGROUND.value))

                data_list.append(one_record)

        return data_list



# Copyright (c) OpenMMLab. All rights reserved.
import random
from typing import List, Optional, Union

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from mmseg.models.utils import resize
from mmseg.utils import (ConfigType, OptConfigType, OptMultiConfig,
                         OptSampleList, SampleList, add_prefix, ForwardResults)

from mmseg.models.segmentors.base import BaseSegmentor
from mmengine.registry import MODELS
import cv2, os


@MODELS.register_module()
class EncoderDecoder(BaseSegmentor):
    """Encoder Decoder segmentors.

    EncoderDecoder typically consists of backbone, decode_head, auxiliary_head.
    Note that auxiliary_head is only used for deep supervision during training,
    which could be dumped during inference.

    1. The ``loss`` method is used to calculate the loss of model,
    which includes two steps: (1) Extracts features to obtain the feature maps
    (2) Call the decode head loss function to forward decode head model and
    calculate losses.

    .. code:: text

     loss(): extract_feat() -> _decode_head_forward_train() -> _auxiliary_head_forward_train (optional)
     _decode_head_forward_train(): decode_head.loss()
     _auxiliary_head_forward_train(): auxiliary_head.loss (optional)

    2. The ``predict`` method is used to predict segmentation results,
    which includes two steps: (1) Run inference function to obtain the list of
    seg_logits (2) Call post-processing function to obtain list of
    ``SegDataSampel`` including ``pred_sem_seg`` and ``seg_logits``.

    .. code:: text

     predict(): inference() -> postprocess_result()
     infercen(): whole_inference()/slide_inference()
     whole_inference()/slide_inference(): encoder_decoder()
     encoder_decoder(): extract_feat() -> decode_head.predict()

    3. The ``_forward`` method is used to output the tensor by running the model,
    which includes two steps: (1) Extracts features to obtain the feature maps
    (2)Call the decode head forward function to forward decode head model.

    .. code:: text

     _forward(): extract_feat() -> _decode_head.forward()

    Args:

        backbone (ConfigType): The config for the backnone of segmentor.
        decode_head (ConfigType): The config for the decode head of segmentor.
        neck (OptConfigType): The config for the neck of segmentor.
            Defaults to None.
        auxiliary_head (OptConfigType): The config for the auxiliary head of
            segmentor. Defaults to None.
        train_cfg (OptConfigType): The config for training. Defaults to None.
        test_cfg (OptConfigType): The config for testing. Defaults to None.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`.
        pretrained (str, optional): The path for pretrained model.
            Defaults to None.
        init_cfg (dict, optional): The weight initialized config for
            :class:`BaseModule`.
        infer_batch_size: 推理时候的Batch大小，默认是1，推荐设置成与训练相同大小的batch
    """  # noqa: E501

    def __init__(self,
                 backbone: ConfigType,
                 decode_head: ConfigType,
                 neck: OptConfigType = None,
                 auxiliary_head: OptConfigType = None,
                 train_cfg: OptConfigType = None,
                 test_cfg: OptConfigType = None,
                 data_preprocessor: OptConfigType = None,
                 pretrained: Optional[str] = None,
                 init_cfg: OptMultiConfig = None,
                 infer_batch_size=1):
        super(EncoderDecoder, self).__init__(
            data_preprocessor=data_preprocessor, init_cfg=init_cfg)
        if pretrained is not None:
            assert backbone.get('pretrained') is None, \
                'both backbone and segmentor set pretrained weight'
            backbone.pretrained = pretrained
        self.backbone = MODELS.build(backbone)
        if neck is not None:
            self.neck = MODELS.build(neck)
        self._init_decode_head(decode_head)
        self._init_auxiliary_head(auxiliary_head)

        self.train_cfg = train_cfg
        self.test_cfg = test_cfg
        self._infer_batch_size = infer_batch_size  # 批量推理计算的时候，一个batch里面有多少小图
        self.trans_stride = 9  # 这个不作为模型外部初始化的参数，只有调用新Loop的时候才改动
        assert self.with_decode_head

    def _init_decode_head(self, decode_head: ConfigType) -> None:
        """Initialize ``decode_head``"""
        self.decode_head = MODELS.build(decode_head)
        self.align_corners = self.decode_head.align_corners
        self.num_classes = self.decode_head.num_classes

    def _init_auxiliary_head(self, auxiliary_head: ConfigType) -> None:
        """Initialize ``auxiliary_head``"""
        if auxiliary_head is not None:
            if isinstance(auxiliary_head, list):
                self.auxiliary_head = nn.ModuleList()
                for head_cfg in auxiliary_head:
                    self.auxiliary_head.append(MODELS.build(head_cfg))
            else:
                self.auxiliary_head = MODELS.build(auxiliary_head)

    def extract_feat(self, inputs: Tensor) -> List[Tensor]:
        """Extract features from images."""
        x = self.backbone(inputs)
        if self.with_neck:
            x = self.neck(x)
        return x

    @staticmethod
    def get_batch_img_metas(inputs, data_samples):
        if data_samples is not None:
            batch_img_metas = [
                data_sample.metainfo for data_sample in data_samples
            ]
        else:
            batch_img_metas = [
                dict(
                    ori_shape=inputs.shape[2:],
                    img_shape=inputs.shape[2:],
                    pad_shape=inputs.shape[2:],
                    padding_size=[0, 0, 0, 0])
            ] * inputs.shape[0]
        return batch_img_metas

    def encode_decode(self, inputs: Tensor,
                      batch_img_metas: List[dict]):
        """Encode images with backbone and decode into a semantic segmentation
        map of the same size as input."""
        x = self.extract_feat(inputs)
        seg_logits = self.decode_head.predict(x, batch_img_metas,
                                              self.test_cfg)

        return seg_logits

    def _decode_head_forward_train(self, inputs: List[Tensor],
                                   data_samples: SampleList) -> dict:
        """Run forward function and calculate loss for decode head in
        training."""
        losses = dict()
        loss_decode = self.decode_head.loss(inputs, data_samples,
                                            self.train_cfg)

        losses.update(add_prefix(loss_decode, 'decode'))
        return losses

    def _auxiliary_head_forward_train(self, inputs: List[Tensor],
                                      data_samples: SampleList) -> dict:
        """Run forward function and calculate loss for auxiliary head in
        training."""
        losses = dict()
        if isinstance(self.auxiliary_head, nn.ModuleList):
            for idx, aux_head in enumerate(self.auxiliary_head):
                loss_aux = aux_head.loss(inputs, data_samples, self.train_cfg)
                losses.update(add_prefix(loss_aux, f'aux_{idx}'))
        else:
            loss_aux = self.auxiliary_head.loss(inputs, data_samples,
                                                self.train_cfg)
            losses.update(add_prefix(loss_aux, 'aux'))

        return losses

    def loss(self, inputs: Tensor, data_samples: SampleList) -> dict:
        """Calculate losses from a batch of inputs and data samples.

        Args:
            inputs (Tensor): Input images.
            data_samples (list[:obj:`SegDataSample`]): The seg data samples.
                It usually includes information such as `metainfo` and
                `gt_sem_seg`.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """

        x = self.extract_feat(inputs)

        losses = dict()

        loss_decode = self._decode_head_forward_train(x, data_samples)
        losses.update(loss_decode)

        if self.with_auxiliary_head:
            loss_aux = self._auxiliary_head_forward_train(x, data_samples)
            losses.update(loss_aux)

        return losses

    def predict(self,
                inputs: Tensor,
                data_samples: OptSampleList = None) -> SampleList:
        """Predict results from a batch of inputs and data samples with post-
        processing.

        Args:
            inputs (Tensor): Inputs with shape (N, C, H, W).
            data_samples (List[:obj:`SegDataSample`], optional): The seg data
                samples. It usually includes information such as `metainfo`
                and `gt_sem_seg`.

        Returns:
            list[:obj:`SegDataSample`]: Segmentation results of the
            input images. Each SegDataSample usually contain:

            - ``pred_sem_seg``(PixelData): Prediction of semantic segmentation.
            - ``seg_logits``(PixelData): Predicted logits of semantic
                segmentation before normalization.
        """
        batch_img_metas = self.get_batch_img_metas(inputs, data_samples)
        seg_logits = self.inference(inputs, batch_img_metas)

        # add background channel when out channel is 1
        # this operation makes the output channels always greater than 1
        if self.decode_head.out_channels == 1:
            seg_logits = torch.cat((1-seg_logits, seg_logits), dim=1)

        return self.postprocess_result(seg_logits, data_samples)

    def _forward(self,
                 inputs: Tensor,
                 data_samples: OptSampleList = None) -> Tensor:
        """Network forward process.

        Args:
            inputs (Tensor): Inputs with shape (N, C, H, W).
            data_samples (List[:obj:`SegDataSample`]): The seg
                data samples. It usually includes information such
                as `metainfo` and `gt_sem_seg`.

        Returns:
            Tensor: Forward output of model without any post-processes.
        """
        x = self.extract_feat(inputs)
        return self.decode_head.forward(x)

    def stability_test(self, inputs: Tensor,  data_samples: OptSampleList = None):
        # 稳定性测试现在是中图级别了,只需要变出来一堆平移和旋转后的中图，然后全部输出检测结果就行，返回的都只有指标了，中图信息太多了
        b, _, h_img, w_img = inputs.size()
        assert b == 1  # 稳定性测试和val一样必须是单样本的模式
        label_map = data_samples[0].gt_sem_seg.data
        if label_map.max() == 0:  # 负样本直接跳过
            return []

        area_list = []
        prob_list = []

        h_stride, w_stride = self.test_cfg.stride

        h_grids = h_stride // self.trans_stride  # 首先，单边pad stride之后就可以遍历所有情况了，超过stride就是下一个周期的
        w_grids = w_stride // self.trans_stride

        for h_index in range(h_grids):
            for w_index in range(w_grids):
                padded = torch.zeros((inputs.size(0),
                                      inputs.size(1),
                                      inputs.size(2) + h_stride,
                                      inputs.size(3) + w_stride)).cuda()
                h_start = h_index * self.trans_stride
                w_start = w_index * self.trans_stride
                padded[:, :, h_start:h_start + h_img, w_start:w_start + w_img] = inputs  # 通过填充的方式实现中图的平移
                batch_img_metas = [dict(ori_shape=padded.shape[2:],
                                        img_shape=padded.shape[2:],
                                        pad_shape=padded.shape[2:],
                                        padding_size=[0, 0, 0, 0])]
                prob_map = self.inference(padded, batch_img_metas)[:, :, h_start:h_start + h_img, w_start:w_start + w_img]
                prob_map = prob_map[:, :, 3:-3, 3:-3]  # 干掉边缘过检
                prob_map, seg_map = prob_map.max(dim=1)
                seg_map = (seg_map != 0) * 1.  # 不管类别，全部按照单类算
                area_list.append(seg_map.sum().item())
                prob_list.append(((seg_map * prob_map).sum() / (seg_map.sum() + 1e-8)).item())
                # name = os.path.basename(data_samples[0].img_path[0])
                # cv2.imwrite(f'/root/data/mmcvlab/lck/workdirs/stable_show/{name}_h{h_index}g{w_index}.png',
                #             seg_map.squeeze().cpu().numpy() * 255)

        return np.array(area_list), np.array(prob_list)

    def forward(self,
                inputs: Tensor,
                data_samples: OptSampleList = None,
                mode: str = 'tensor',
                kwargs: dict = None) -> ForwardResults:
        """The unified entry for a forward process in both training and test.
        The method should accept three modes: "tensor", "predict" and "loss":
        - "tensor": Forward the whole network and return tensor or tuple of
        tensor without any post-processing, same as a common nn.Module.
        - "predict": Forward and return the predictions, which are fully
        processed to a list of :obj:`SegDataSample`.
        - "loss": Forward and return a dict of losses according to the given
        inputs and data samples.
        Note that this method doesn't handle neither back propagation nor
        optimizer updating, which are done in the :meth:`train_step`.
        Args:
            inputs (torch.Tensor): The input tensor with shape (N, C, ...) in
                general.
            data_samples (list[:obj:`SegDataSample`]): The seg data samples.
                It usually includes information such as `metainfo` and
                `gt_sem_seg`. Default to None.
            mode (str): Return what kind of value. Defaults to 'tensor'.
            kwargs (dict): 一系列用在deploy过程中的参数，主要包括：
                    1. output_mode: 'raw' / 'argmax'
                        raw: 直接返回原始模型输出，不过带了sigmoid
                        argmax: 会在sigmoid的基础上，进行argmax，返回label+置信度图
                    2. threshold: 二值化阈值
        Returns:
            The return type depends on ``mode``.
            - If ``mode="tensor"``, return a tensor or a tuple of tensor.
            - If ``mode="predict"``, return a list of :obj:`DetDataSample`.
            - If ``mode="loss"``, return a dict of tensor.
            - If ``mode="deploy"``, forward for deployment.
        """
        if mode == 'loss':
            return self.loss(inputs, data_samples)
        elif mode == 'predict':
            return self.predict(inputs, data_samples)
        elif mode == 'tensor':
            return self._forward(inputs, data_samples)
        elif mode == 'deploy':
            if kwargs:
                return self.deploy(inputs, **kwargs)
            else:
                return self.deploy(inputs)
        elif mode == 'stability_test':
            return self.stability_test(inputs, data_samples)
        else:
            raise RuntimeError(f'Invalid mode "{mode}". '
                               'Only supports loss, predict and tensor mode')

    def deploy(self,
               inputs: Tensor,
               output_mode='raw'):
        """
        目前输出模式支持两种：
        “raw”：
        “argmax”：
        """
        height, width = inputs.shape[-2:]
        # deploy mode 的 preprocessor
        x = self.data_preprocessor.deploy(inputs)
        # 网络结构的正常 forward
        x = self._forward(x)

        # when height&width of output is not consistent with the input, resize for consistency
        out_height, out_width = x.shape[-2:]
        if out_height != height and out_width != width:
            x = resize(
                input=x,
                size=(height, width),
                mode='bilinear',
                align_corners=self.align_corners
            )

        # when only one channel for prediction, sigmoid_before_loss must be true
        if self.decode_head.out_channels == 1:
            assert self.decode_head.sigmoid_before_loss, 'decode_head.sigmoid_before_loss must be true if prediction channel is 1.'
            x = x.sigmoid()
            x = torch.cat((1-x, x), dim=1)  # add background score to the first channel
        else:
            if self.decode_head.sigmoid_before_loss:
                x = x.sigmoid()
            else:
                x = x.softmax(dim=1)

        if output_mode == 'raw':
            return x
        elif 'argmax' in output_mode:
            scores, indexes = x.max(dim=1, keepdim=True)
            indexes = indexes.to(torch.uint8) if 'int8' in output_mode else indexes.to(torch.float32)
            return scores, indexes
        elif 'onlylabel' in output_mode:
            _, indexes = x.max(dim=1, keepdim=True)
            indexes = indexes.to(torch.uint8) if 'int8' in output_mode else indexes.to(torch.float32)
            return indexes
        else:
            raise ValueError(f'Unsupported output_mode: {output_mode}.')

    @staticmethod
    def _crop_region_in_tensor(img, x, y, region_size, segmap_size):  # 根据随机出来的y和x进行region的裁剪
        """根据随机出来的x和y进行region的裁剪"""
        region_h, region_w = region_size  # 需要切分的小图的分辨率
        segmap_h, segmap_w = segmap_size  # 分割结果segmap的分辨率

        # pad_w 计算的是水平方向上，Crop出来的小图和Gt的尺寸的差异
        pad_w = int((region_w - segmap_w) / 2.)
        # pad_h 计算的是垂直方向上，Crop出来的小图和Gt的尺寸的差异
        pad_h = int((region_h - segmap_h) / 2.)

        # 这个是根据x和pad_w计算图像左边需要pad的宽度，当x > pad_w的时候，意味着左边不用pad,就得到的0，不然就需要pad
        pad_w_l = max(0, pad_w - x)
        # 这个是根据x和pad_w计算图像左边需要pad的宽度，
        pad_w_r = max(0, x - pad_w + region_w - img.size(3))
        # 高度方向的逻辑同上
        pad_h_t = max(0, pad_h - y)
        pad_h_b = max(0, y - pad_h + region_h - img.size(2))

        tmp = F.pad(img, [pad_w_l, pad_w_r, pad_h_t, pad_h_b])

        x_ = x - pad_w + pad_w_l  # x, y 都是segmap分割结果在原图坐标系下的坐标
        y_ = y - pad_h + pad_h_t  # x_, y_ 就是region在pad后的图像下的坐标，因此需要减掉pad再加上在原图左上角pad的大小
        region = tmp[:, :, y_:y_ + region_h, x_:x_ + region_w]
        return region

    def slide_inference(self,
                        inputs: Tensor,
                        batch_img_metas: List[dict],
                        ) -> Tensor:
        """Inference by sliding-window with overlap.

        If h_crop > h_img or w_crop > w_img, the small patch will be used to
        decode without padding.

        Args:
            inputs (tensor): the tensor should have a shape NxCxHxW,
                which contains all images in the batch.
            batch_img_metas (List[dict]): List of image metainfo where each may
                also contain: 'img_shape', 'scale_factor', 'flip', 'img_path',
                'ori_shape', and 'pad_shape'.
                For details on the values of these keys see
                `mmseg/datasets/pipelines/formatting.py:PackSegInputs`.

        Returns:
            Tensor: The segmentation results, seg_logits from model of each
                input image.
        """
        h_stride, w_stride = self.test_cfg.stride
        label_h, label_w = self.test_cfg.crop_size
        batch_size, _, h_img, w_img = inputs.size()
        num_head_out_channels = self.decode_head.out_channels
        assert h_img >= self.test_cfg.crop_size[0]
        assert w_img >= self.test_cfg.crop_size[1]
        # # 测试阶段由于前面的preprocessor已经不会有更小的图了
        # pad_h_b = int(max(0, label_h - original_h_img))
        # pad_w_r = int(max(0, label_w - original_w_img))
        # inputs = F.pad(inputs, [0, pad_w_r, 0, pad_h_b])
        # batch_size, _, h_img, w_img = inputs.size()  # 重新计算尺寸

        # 准备preds用来保留预测结果
        preds = inputs.new_zeros((batch_size, num_head_out_channels, h_img, w_img))
        # count_mat是保留每个像素点的计算次数
        count_mat = inputs.new_zeros((batch_size, 1, h_img, w_img))

        # 计算水平和垂直方向要切分几次
        h_grids = int(max(h_img - label_h + h_stride - 1, 0) // h_stride + 1)
        w_grids = int(max(w_img - label_w + w_stride - 1, 0) // w_stride + 1)

        # 计算一共要切几个小图，方便和前面指定的batch_size进行比较
        grid_num = h_grids * w_grids

        # 把这些小图，按照batsize的大小分开，分批进行推理，这边生成的grid_batches里面存的就是每一次批量推理时候，这批小图的id
        grid_batches = [range(i, min(i+self._infer_batch_size, grid_num))
                        for i in range(0, grid_num, self._infer_batch_size)]

        for grid_batch in grid_batches:  # grid_batch就是这批小图的空间id
            crop_imgs = []  # 用来存储切出来的小图
            crop_locs = []  # 用来存储小图的坐标信息
            for grid_idx in grid_batch:
                h_idx = grid_idx // w_grids
                w_idx = grid_idx % w_grids
                y1 = h_idx * h_stride  # x1 y1 是初步的左上角点
                x1 = w_idx * w_stride
                y2 = int(min(y1 + label_h, h_img))  # x2和y2是计算之后得到的右下角点
                x2 = int(min(x1 + label_w, w_img))
                y1 = int(max(y2 - label_h, 0))  # 再算一次是因为可能x2和y2存在越界的可能
                x1 = int(max(x2 - label_w, 0))  # 所以x1和y1可能比初步计算的结果要小一点
                crop_img = self._crop_region_in_tensor(inputs, x1, y1,
                                                       region_size=self.test_cfg.crop_size,
                                                       segmap_size=(label_h, label_w))
                crop_imgs.append(crop_img)  # 先把小图存起来
                crop_locs.append([x1, y1])  # 对应裁剪的位置也存起来

            crop_imgs = torch.cat(crop_imgs, dim=0)  # 在Batch维度把他拼起来
            # change the image shape to patch shape
            batch_img_metas[0]['img_shape'] = crop_imgs.shape[2:]   # 原版
            batch_img_metas[0]['pad_shape'] = crop_imgs.shape[2:]   # 原版
            crop_seg_logit = self.encode_decode(crop_imgs, batch_img_metas)

            for i in range(len(crop_locs)):
                xi, yi = crop_locs[i]
                if self.decode_head.out_channels == 1:
                    assert self.decode_head.sigmoid_before_loss, \
                        'decode_head.sigmoid_before_loss must be true if prediction channel is 1.'
                    preds[:, :, yi:yi + label_h, xi:xi + label_w] += crop_seg_logit[i].sigmoid()
                else:
                    if self.decode_head.sigmoid_before_loss:
                        preds[:, :, yi:yi + label_h, xi:xi + label_w] += crop_seg_logit[i].sigmoid()
                    else:
                        preds[:, :, yi:yi + label_h, xi:xi + label_w] += crop_seg_logit[i].softmax(dim=0)
                count_mat[:, :, yi:yi + label_h, xi:xi + label_w] += 1

        assert (count_mat == 0).sum() == 0
        seg_logits = preds / count_mat

        return seg_logits

    def center_slide_inference(self, inputs: Tensor, batch_img_metas: List[dict]) -> Tensor:
        """Inference by sliding-window without overlap, by just save the center result of segmentation
        Args:
            inputs (tensor): the tensor should have a shape NxCxHxW,
                which contains all images in the batch.
            batch_img_metas (List[dict]): List of image meta info where each may
                also contain: 'img_shape', 'scale_factor', 'flip', 'img_path',
                'ori_shape', and 'pad_shape'.
                For details on the values of these keys see
                `mmseg/datasets/pipelines/formatting.py:PackSegInputs`.

        Returns:
            Tensor: The segmentation results, seg_logit from model of each
                input image.
        """
        # Need to pad the inputs before inference
        # h_stride, w_stride = self.test_cfg.stride
        label_h, label_w = self.test_cfg.crop_size
        pad_h, pad_w = int(label_h / 8), int(label_w / 8)
        # the stride is deside by pad_out, not manually set
        h_stride, w_stride = label_h - 2 * pad_h, label_w - 2 * pad_w

        batch_size, _, h_img, w_img = inputs.size()
        num_head_out_channels = self.decode_head.out_channels
        assert h_img >= self.test_cfg.crop_size[0]
        assert w_img >= self.test_cfg.crop_size[1]

        inputs = F.pad(inputs, [pad_w, pad_w, pad_h, pad_h], value=-1)
        batch_size, _, h_img, w_img = inputs.size()  # 重新计算尺寸

        # 准备pred用来保留预测结果
        pred = inputs.new_zeros((batch_size, num_head_out_channels, h_img, w_img))
        # count_mat是保留每个像素点的计算次数
        count_mat = inputs.new_zeros((batch_size, 1, h_img, w_img))

        # 计算水平和垂直方向要切分几次
        h_grids = int(max(h_img - label_h + h_stride - 1, 0) // h_stride + 1)
        w_grids = int(max(w_img - label_w + w_stride - 1, 0) // w_stride + 1)

        # 计算一共要切几个小图，方便和前面指定的batch_size进行比较
        grid_num = h_grids * w_grids

        # 把这些小图，按照batch size的大小分开，分批进行推理，这边生成的grid_batches里面存的就是每一次批量推理时候，这批小图的id
        grid_batches = [range(i, min(i+self._infer_batch_size, grid_num))
                        for i in range(0, grid_num, self._infer_batch_size)]

        for grid_batch in grid_batches:  # grid_batch就是这批小图的空间id
            crop_img_list = []  # 用来存储切出来的小图
            crop_locs = []  # 用来存储小图的坐标信息
            for grid_idx in grid_batch:
                h_idx = grid_idx // w_grids
                w_idx = grid_idx % w_grids
                y1 = h_idx * h_stride  # x1 y1 是初步的左上角点
                x1 = w_idx * w_stride
                y2 = int(min(y1 + label_h, h_img))  # x2和y2是计算之后得到的右下角点
                x2 = int(min(x1 + label_w, w_img))
                y1 = int(max(y2 - label_h, 0))  # 再算一次是因为可能x2和y2存在越界的可能
                x1 = int(max(x2 - label_w, 0))  # 所以x1和y1可能比初步计算的结果要小一点
                crop_img = self._crop_region_in_tensor(inputs, x1, y1,
                                                       region_size=self.test_cfg.crop_size,
                                                       segmap_size=(label_h, label_w))
                crop_img_list.append(crop_img)  # 先把小图存起来
                crop_locs.append([x1, y1])  # 对应裁剪的位置也存起来

            crop_img_list = torch.cat(crop_img_list, dim=0)  # 在Batch维度把他拼起来
            # change the image shape to patch shape
            batch_img_metas[0]['img_shape'] = crop_img_list.shape[2:]   # 原版
            batch_img_metas[0]['pad_shape'] = crop_img_list.shape[2:]   # 原版
            crop_seg_logit = self.encode_decode(crop_img_list, batch_img_metas)[:, :, pad_h:-pad_h, pad_w:-pad_w]

            for i in range(len(crop_locs)):
                xi, yi = crop_locs[i]
                if self.decode_head.out_channels == 1:
                    assert self.decode_head.sigmoid_before_loss, \
                        'decode_head.sigmoid_before_loss must be true if prediction channel is 1.'
                    pred[:, :, yi+pad_h:yi+label_h-pad_h, xi+pad_w:xi+label_w-pad_w] += crop_seg_logit[i].sigmoid()
                else:
                    if self.decode_head.sigmoid_before_loss:
                        pred[:, :, yi+pad_h:yi+label_h-pad_h, xi+pad_w:xi+label_w-pad_w] += crop_seg_logit[i].sigmoid()
                    else:
                        pred[:, :, yi+pad_h:yi+label_h-pad_h, xi+pad_w:xi+label_w-pad_w] += crop_seg_logit[i].softmax(dim=0)
                count_mat[:, :, yi+pad_h:yi+label_h-pad_h, xi+pad_w:xi+label_w-pad_w] += 1
        count_mat = count_mat[:, :, pad_h:-pad_h, pad_w:-pad_w]
        pred = pred[:, :, pad_h:-pad_h, pad_w:-pad_w]
        assert (count_mat == 0).sum() == 0
        seg_logit = pred / count_mat

        return seg_logit

    def whole_inference(self, inputs: Tensor,
                        batch_img_metas: List[dict]) -> Tensor:
        """Inference with full image.

        Args:
            inputs (Tensor): The tensor should have a shape NxCxHxW, which
                contains all images in the batch.
            batch_img_metas (List[dict]): List of image metainfo where each may
                also contain: 'img_shape', 'scale_factor', 'flip', 'img_path',
                'ori_shape', and 'pad_shape'.
                For details on the values of these keys see
                `mmseg/datasets/pipelines/formatting.py:PackSegInputs`.

        Returns:
            Tensor: The segmentation results, seg_logits from model of each
                input image.
        """

        seg_logits = self.encode_decode(inputs, batch_img_metas)
        if self.decode_head.out_channels == 1:
            assert self.decode_head.sigmoid_before_loss, 'decode_head.sigmoid_before_loss must be true if prediction channel is 1.'
            seg_logits = seg_logits.sigmoid()
        else:
            if self.decode_head.sigmoid_before_loss:
                seg_logits = seg_logits.sigmoid()
            else:
                seg_logits = seg_logits.softmax(dim=1)
        return seg_logits

    def inference(self, inputs: Tensor, batch_img_metas: List[dict]) -> Tensor:
        """Inference with slide/whole style.

        Args:
            inputs (Tensor): The input image of shape (N, 3, H, W).
            batch_img_metas (List[dict]): List of image metainfo where each may
                also contain: 'img_shape', 'scale_factor', 'flip', 'img_path',
                'ori_shape', 'pad_shape', and 'padding_size'.
                For details on the values of these keys see
                `mmseg/datasets/pipelines/formatting.py:PackSegInputs`.

        Returns:
            Tensor: The segmentation results, seg_logits from model of each
                input image.
        """

        assert self.test_cfg.mode in ['slide', 'whole', 'center_slide']
        ori_shape = batch_img_metas[0]['ori_shape']
        assert all(_['ori_shape'] == ori_shape for _ in batch_img_metas)
        if self.test_cfg.mode == 'slide':
            # 普通滑窗模式的推理
            seg_logit = self.slide_inference(inputs, batch_img_metas)
        elif self.test_cfg.mode == 'center_slide':
            seg_logit = self.center_slide_inference(inputs, batch_img_metas)
        else:
            seg_logit = self.whole_inference(inputs, batch_img_metas)
        # 这两个新的inference里边，已经完成了激活的操作，和deploy时候的顺序是一致的
        return seg_logit

    def stability_test_step(self, data: Union[tuple, dict, list], trans_stride: int) -> list:
        self.trans_stride = trans_stride
        data = self.data_preprocessor(data, False)
        return self._run_forward(data, mode='stability_test')



import argparse
import logging
import os
import os.path as osp
import platform
import sys
import json
from typing import Union, List, Dict, Any

from mmdet.utils import setup_cache_size_limit_of_dynamo
from mmengine.config import Config, DictAction
from mmengine.logging import print_log
from mmengine.runner import Runner
from mmengine.registry import RUNNERS
from mmyolo.utils import is_metainfo_lower

try:
    import visaionlibrary
except ImportError:
    raise ImportError("visaionlibrary not found, please check your environment")

def _extract_baseParam_visaion_keys(data, result=None):
    """
    递归提取以 VISAION 开头的 key 和 value
    """
    if result is None:
        result = {}
    if isinstance(data, dict):
        for key, value in data.items():
            if key.startswith('VISAION'):
                result[key] = value
            _extract_baseParam_visaion_keys(value, result)
    elif isinstance(data, list):
        for item in data:
            _extract_baseParam_visaion_keys(item, result)
    return result

def _extract_page_visaion_keys(data, result=None):
    """
    递归提取以 VISAION 开头的 key 和 value
    """
    if result is None:
        result = {}
    if isinstance(data, dict):
        if 'key' in data and 'value' in data:
            if data['key'].startswith('VISAION'):
                result[data['key']] = data['value']
        for _, value in data.items():
            _extract_page_visaion_keys(value, result)
    elif isinstance(data, list):
        for item in data:
            _extract_page_visaion_keys(item, result)
    return result

def _update_meta_info(meta_info_origin: Union[Dict[str, Any], List[Dict[str, Any]]], task_type: str) -> Dict[str, Any]:
    if task_type == 'det':
        classes = [tmp['name'] for tmp in meta_info_origin]
        palette = [tuple(tmp['color'][:3]) for tmp in meta_info_origin]
        meta_info = {"classes": classes, "palette": palette}
        return meta_info
    elif task_type == 'seg':
        classes = ['background'] + [tmp['name'] for tmp in meta_info_origin]
        palette = [(0, 0, 0)] + [tuple(tmp['color'][:3]) for tmp in meta_info_origin]
        meta_info = {"classes": classes, "palette": palette}
        return meta_info
    elif task_type == 'cls':
        classes = ['background'] + [tmp['name'] for tmp in meta_info_origin]
        palette = [(0, 0, 0)] + [tuple(tmp['color'][:3]) for tmp in meta_info_origin]
        meta_info = {"classes": classes, "palette": palette}
        return meta_info
    elif task_type == 'ins':
        classes = [tmp['name'] for tmp in meta_info_origin]
        palette = [tuple(tmp['color'][:3]) for tmp in meta_info_origin]
        meta_info = {"classes": classes, "palette": palette}
        return meta_info
    else:
        raise ValueError(f"task_type {task_type} not supported")

def parse_args():
    parser = argparse.ArgumentParser(description='Train a model')
    parser.add_argument('template', help='train config file path')
    parser.add_argument(
        '--launcher',
        choices=['none', 'pytorch', 'slurm', 'mpi'],
        default='none',
        help='job launcher')
    # When using PyTorch version >= 2.0.0, the `torch.distributed.launch`
    # will pass the `--local-rank` parameter to `tools/train.py` instead
    # of `--local_rank`.
    parser.add_argument('--local_rank', '--local-rank', type=int, default=0)
    args = parser.parse_args()
    if 'LOCAL_RANK' not in os.environ:
        os.environ['LOCAL_RANK'] = str(args.local_rank)

    return args

def train():
    args = parse_args()

    # 解析参数
    with open(args.template, 'r', encoding='utf-8') as f:
        job_info = json.load(f)
    
    # 提取project_id, job_id, job_name, job_args
    project_id = job_info['project_id']     # 项目id
    job_id = job_info['job_id']             # 任务id
    job_type = job_info['job_type']         # 任务名称
    job_args = job_info['job_args']           # 任务参数
    assert project_id is not None, "project_id is not None"
    assert job_id is not None, "job_id is not None"
    assert job_type is not None, "job_type is not None"
    assert job_args is not None, "job_args is not None"

    # 提取以 VISAION 开头的 key 和 value
    visaion_base_params = _extract_baseParam_visaion_keys(job_args['baseParam'])    # 提取基础参数
    visaion_page_params = _extract_page_visaion_keys(job_args['page'])              # 提取页面参数
    visaion_params = {**visaion_base_params, **visaion_page_params}                 # 合并参数
    # 获取VISAION_DIR
    VISAION_DIR = visaion_params['VISAION_DIR']
    # 更新VISAION_META_INFO
    if 'VISAION_META_INFO' in visaion_params:   
        visaion_params['VISAION_META_INFO'] = _update_meta_info(visaion_params['VISAION_META_INFO'], visaion_params['VISAION_TEMPLATE_TYPE'])
    # 更新VISAION_CHECKPOINT
    if "VISAION_CHECKPOINT" in visaion_params and visaion_params["VISAION_CHECKPOINT"]:  
        visaion_params["VISAION_CHECKPOINT"] = osp.join(VISAION_DIR, 'weights', visaion_params["VISAION_CHECKPOINT"]).replace('\\', '/')
    # 更新VISAION_LOAD_FROM
    if "VISAION_LOAD_FROM" in visaion_params and visaion_params["VISAION_LOAD_FROM"]:  
        visaion_params["VISAION_LOAD_FROM"] = osp.join(VISAION_DIR, 'weights', visaion_params["VISAION_LOAD_FROM"]).replace('\\', '/')
    # 更新VISAION_NUM_CLASSES
    visaion_params['VISAION_NUM_CLASSES'] = len(visaion_params['VISAION_META_INFO']['classes'])
    # 更新VISAION_BACKEND
    if platform.system() == 'Windows':  
        visaion_params['VISAION_BACKEND'] = 'gloo'
    # 更新VISAION_IN_CHANNELS
    if visaion_params["VISAION_COLOR_TYPE"] == "color": 
        visaion_params["VISAION_IN_CHANNELS"] = 3
    elif visaion_params["VISAION_COLOR_TYPE"] == "grayscale":
        visaion_params["VISAION_IN_CHANNELS"] = 1
    elif visaion_params["VISAION_COLOR_TYPE"] == "gray" or visaion_params["VISAION_COLOR_TYPE"] == "grey":
        print_log(f"VISAION_COLOR_TYPE {visaion_params['VISAION_COLOR_TYPE']} is deprecated, use grayscale instead")
        visaion_params["VISAION_IN_CHANNELS"] = 1
        visaion_params["VISAION_COLOR_TYPE"] = 'grayscale'
    else:
        raise ValueError(f"VISAION_COLOR_TYPE {visaion_params['VISAION_COLOR_TYPE']} not supported")
    # 更新VISAION_TRAIN_DATA_ROOT
    visaion_train_data_root = []
    if "VISAION_TRAIN_DATA_ROOT" in visaion_params and visaion_params["VISAION_TRAIN_DATA_ROOT"]:
        for data_name in visaion_params["VISAION_TRAIN_DATA_ROOT"]:
            visaion_train_data_root.append(osp.join(VISAION_DIR, "projects", str(project_id), "datasets", data_name).replace('\\', '/'))
    visaion_params["VISAION_TRAIN_DATA_ROOT"] = "::".join(visaion_train_data_root)
    # 更新VISAION_VAL_DATA_ROOT
    visaion_val_data_root = []
    if "VISAION_VAL_DATA_ROOT" in visaion_params and visaion_params["VISAION_VAL_DATA_ROOT"]:
        for data_name in visaion_params["VISAION_VAL_DATA_ROOT"]:
            visaion_val_data_root.append(osp.join(VISAION_DIR, "projects", str(project_id), "datasets", data_name).replace('\\', '/'))
    visaion_params["VISAION_VAL_DATA_ROOT"] = "::".join(visaion_val_data_root)
    # 更新VISAION_TEST_DATA_ROOT
    visaion_test_data_root = []
    if "VISAION_TEST_DATA_ROOT" in visaion_params and visaion_params["VISAION_TEST_DATA_ROOT"]:
        for data_name in visaion_params["VISAION_TEST_DATA_ROOT"]:
            visaion_test_data_root.append(osp.join(VISAION_DIR, "projects", str(project_id), "datasets", data_name).replace('\\', '/'))
    visaion_params["VISAION_TEST_DATA_ROOT"] = "::".join(visaion_test_data_root)
    # 更新VISAION_TEMPLATE_NAME
    visaion_params['VISAION_TEMPLATE_NAME'] = osp.join(VISAION_DIR, 'templates', visaion_params['VISAION_TEMPLATE_NAME']).replace('\\', '/')
    assert osp.exists(visaion_params['VISAION_TEMPLATE_NAME']), f"VISAION_TEMPLATE_NAME {visaion_params['VISAION_TEMPLATE_NAME']} not exists"
    # 更新VISAION_WORK_DIR
    visaion_params['VISAION_WORK_DIR'] = osp.join(VISAION_DIR, 'projects', str(project_id), job_type, str(job_id)).replace('\\', '/')
    os.makedirs(visaion_params['VISAION_WORK_DIR'], exist_ok=True)
    # 更新VISAION_SCALE_FACTOR
    if "VISAION_SCALE_FACTOR" in visaion_params:
        visaion_params['VISAION_SCALE_FACTOR'] = float(visaion_params['VISAION_SCALE_FACTOR'])
    #  打印 VISAION 参数, 设置环境变量
    print_log('setting environment variables...')
    for k, v in visaion_params.items():
        v = str(v)
        # windows 下使用\\加载config时会失败, 所以需要替换为/
        if platform.system() == 'Windows':
            v = v.replace('\\', '/')
        print_log(f"{k}={v}")
        os.environ[k] = v
    
    # Reduce the number of repeated compilations and improve
    # training speed.
    setup_cache_size_limit_of_dynamo()

    # load config, and update configs
    cfg = Config.fromfile(visaion_params['VISAION_TEMPLATE_NAME'] )
    cfg.work_dir = visaion_params['VISAION_WORK_DIR']
    cfg.launcher = args.launcher
    
    # 打印visaionlibrary版本
    print_log(f"visaionlibrary version: {visaionlibrary.__version__}")
    
    # enable automatic-mixed-precision training
    if "VISAION_AMP" in visaion_params and visaion_params['VISAION_AMP'] is True:
        optim_wrapper = cfg.optim_wrapper.type
        if optim_wrapper == 'AmpOptimWrapper':
            print_log(
                'AMP training is already enabled in your config.',
                logger='current',
                level=logging.WARNING)
        else:
            assert optim_wrapper == 'OptimWrapper', (
                '`--amp` is only supported when the optimizer wrapper type is '
                f'`OptimWrapper` but got {optim_wrapper}.')
            cfg.optim_wrapper.type = 'AmpOptimWrapper'
            cfg.optim_wrapper.loss_scale = 'dynamic'

    # resume is determined in this priority: resume from > auto_resume
    if visaion_params.get('VISAION_CHECKPOINT', None):
        cfg.load_from = visaion_params['VISAION_LOAD_FROM']

    # Determine whether the custom metainfo fields are all lowercase
    is_metainfo_lower(cfg)

    # build the runner from config
    if 'runner_type' not in cfg:
        # build the default runner
        runner = Runner.from_cfg(cfg)
    else:
        # build customized runner from the registry
        # if 'runner_type' is set in the cfg
        runner = RUNNERS.build(cfg)

    # start training
    runner.train()

    runner.logger.info('VISAION_DONE')

if __name__ == '__main__':
    train()



# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Optional

import numpy as np
from mmengine.hooks import ParamSchedulerHook
from mmengine.runner import Runner

from mmengine.registry import HOOKS


def linear_fn(lr_factor: float, max_iters: int):
    """Generate linear function."""
    return lambda x: (1 - x / max_iters) * (1.0 - lr_factor) + lr_factor


def cosine_fn(lr_factor: float, max_iters: int):
    """Generate cosine function."""
    return lambda x: (
        (1 - math.cos(x * math.pi / max_iters)) / 2) * (lr_factor - 1) + 1


@HOOKS.register_module()
class YOLOv5IterParamSchedulerHook(ParamSchedulerHook):
    """A hook to update learning rate and momentum in optimizer of YOLOv5."""
    priority = 9

    scheduler_maps = {'linear': linear_fn, 'cosine': cosine_fn}

    def __init__(self,
                 scheduler_type: str = 'linear',
                 lr_factor: float = 0.01,
                 max_iters: int = 300,
                 warmup_iters: int = 60,
                 warmup_bias_lr: float = 0.1,
                 warmup_momentum: float = 0.8,
                 warmup_mim_iter: int = 1000,
                 **kwargs):

        assert scheduler_type in self.scheduler_maps

        self.warmup_iters = warmup_iters
        self.warmup_bias_lr = warmup_bias_lr
        self.warmup_momentum = warmup_momentum
        self.warmup_mim_iter = warmup_mim_iter

        kwargs.update({'lr_factor': lr_factor, 'max_iters': max_iters})
        self.scheduler_fn = self.scheduler_maps[scheduler_type](**kwargs)

        self._base_lr = None
        self._base_momentum = None

    def before_train(self, runner: Runner):
        """Operations before train.

        Args:
            runner (Runner): The runner of the training process.
        """
        optimizer = runner.optim_wrapper.optimizer
        for group in optimizer.param_groups:
            # If the param is never be scheduled, record the current value
            # as the initial value.
            group.setdefault('initial_lr', group['lr'])
            group.setdefault('initial_momentum', group.get('momentum', -1))

        self._base_lr = [
            group['initial_lr'] for group in optimizer.param_groups
        ]
        self._base_momentum = [
            group['initial_momentum'] for group in optimizer.param_groups
        ]

    def before_train_iter(self,
                          runner: Runner,
                          batch_idx: int,
                          data_batch: Optional[dict] = None):
        """Operations before each training iteration.

        Args:
            runner (Runner): The runner of the training process.
            batch_idx (int): The index of the current batch in the train loop.
            data_batch (dict or tuple or list, optional): Data from dataloader.
        """
        cur_iters = runner.iter
        optimizer = runner.optim_wrapper.optimizer

        # The minimum warmup is self.warmup_mim_iter
        warmup_total_iters = max(self.warmup_iters, self.warmup_mim_iter)

        if cur_iters <= warmup_total_iters:
            xp = [0, warmup_total_iters]
            for group_idx, param in enumerate(optimizer.param_groups):
                if group_idx == 2:
                    # bias learning rate will be handled specially
                    yp = [
                        self.warmup_bias_lr,
                        self._base_lr[group_idx] * self.scheduler_fn(cur_iters)
                    ]
                else:
                    yp = [
                        0.0,
                        self._base_lr[group_idx] * self.scheduler_fn(cur_iters)
                    ]
                param['lr'] = np.interp(cur_iters, xp, yp)

                if 'momentum' in param:
                    param['momentum'] = np.interp(
                        cur_iters, xp,
                        [self.warmup_momentum, self._base_momentum[group_idx]])
        else:
            optimizer = runner.optim_wrapper.optimizer
            for group_idx, param in enumerate(optimizer.param_groups):
                param['lr'] = self._base_lr[group_idx] * self.scheduler_fn(cur_iters)




import glob
import math
import os
import re
import shutil
from importlib import metadata
from pathlib import Path

import torch

from visaionlibrary.utils.yoloe.utils import (
    LINUX,
    MACOS,
    PYTHON_VERSION,
    ROOT,
    WINDOWS,
    downloads,
    emojis,
    url2file,
)


def parse_version(version="0.0.0") -> tuple:
    """
    Convert a version string to a tuple of integers, ignoring any extra non-numeric string attached to the version. This
    function replaces deprecated 'pkg_resources.parse_version(v)'.

    Args:
        version (str): Version string, i.e. '2.0.1+cpu'

    Returns:
        (tuple): Tuple of integers representing the numeric part of the version and the extra string, i.e. (2, 0, 1)
    """
    try:
        return tuple(map(int, re.findall(r"\d+", version)[:3]))  # '2.0.1+cpu' -> (2, 0, 1)
    except Exception as e:
        # LOGGER.warning(f"WARNING ⚠️ failure for parse_version({version}), returning (0, 0, 0): {e}")
        pass
        return 0, 0, 0


def is_ascii(s) -> bool:
    """
    Check if a string is composed of only ASCII characters.

    Args:
        s (str): String to be checked.

    Returns:
        (bool): True if the string is composed only of ASCII characters, False otherwise.
    """
    # Convert list, tuple, None, etc. to string
    s = str(s)

    # Check if the string is composed of only ASCII characters
    return all(ord(c) < 128 for c in s)


def check_imgsz(imgsz, stride=32, min_dim=1, max_dim=2, floor=0):
    """
    Verify image size is a multiple of the given stride in each dimension. If the image size is not a multiple of the
    stride, update it to the nearest multiple of the stride that is greater than or equal to the given floor value.

    Args:
        imgsz (int | cList[int]): Image size.
        stride (int): Stride value.
        min_dim (int): Minimum number of dimensions.
        max_dim (int): Maximum number of dimensions.
        floor (int): Minimum allowed value for image size.

    Returns:
        (List[int]): Updated image size.
    """
    # Convert stride to integer if it is a tensor
    stride = int(stride.max() if isinstance(stride, torch.Tensor) else stride)

    # Convert image size to list if it is an integer
    if isinstance(imgsz, int):
        imgsz = [imgsz]
    elif isinstance(imgsz, (list, tuple)):
        imgsz = list(imgsz)
    elif isinstance(imgsz, str):  # i.e. '640' or '[640,640]'
        imgsz = [int(imgsz)] if imgsz.isnumeric() else eval(imgsz)
    else:
        raise TypeError(
            f"'imgsz={imgsz}' is of invalid type {type(imgsz).__name__}. "
            f"Valid imgsz types are int i.e. 'imgsz=640' or list i.e. 'imgsz=[640,640]'"
        )

    # Apply max_dim
    if len(imgsz) > max_dim:
        msg = (
            "'train' and 'val' imgsz must be an integer, while 'predict' and 'export' imgsz may be a [h, w] list "
            "or an integer, i.e. 'yolo export imgsz=640,480' or 'yolo export imgsz=640'"
        )
        if max_dim != 1:
            raise ValueError(f"imgsz={imgsz} is not a valid image size. {msg}")
        # LOGGER.warning(f"WARNING ⚠️ updating to 'imgsz={max(imgsz)}'. {msg}")
        pass
        imgsz = [max(imgsz)]
    # Make image size a multiple of the stride
    sz = [max(math.ceil(x / stride) * stride, floor) for x in imgsz]

    # Print warning message if image size was updated
    if sz != imgsz:
        # LOGGER.warning(f"WARNING ⚠️ imgsz={imgsz} must be multiple of max stride {stride}, updating to {sz}")
        pass
        pass

    # Add missing dimensions if necessary
    sz = [sz[0], sz[0]] if min_dim == 2 and len(sz) == 1 else sz[0] if min_dim == 1 and len(sz) == 1 else sz

    return sz


def check_version(
    current: str = "0.0.0",
    required: str = "0.0.0",
    name: str = "version",
    hard: bool = False,
    verbose: bool = False,
    msg: str = "",
) -> bool:
    """
    Check current version against the required version or range.

    Args:
        current (str): Current version or package name to get version from.
        required (str): Required version or range (in pip-style format).
        name (str, optional): Name to be used in warning message.
        hard (bool, optional): If True, raise an AssertionError if the requirement is not met.
        verbose (bool, optional): If True, print warning message if requirement is not met.
        msg (str, optional): Extra message to display if verbose.

    Returns:
        (bool): True if requirement is met, False otherwise.

    Example:
        ```python
        # Check if current version is exactly 22.04
        check_version(current="22.04", required="==22.04")

        # Check if current version is greater than or equal to 22.04
        check_version(current="22.10", required="22.04")  # assumes '>=' inequality if none passed

        # Check if current version is less than or equal to 22.04
        check_version(current="22.04", required="<=22.04")

        # Check if current version is between 20.04 (inclusive) and 22.04 (exclusive)
        check_version(current="21.10", required=">20.04,<22.04")
        ```
    """
    if not current:  # if current is '' or None
        # LOGGER.warning(f"WARNING ⚠️ invalid check_version({current}, {required}) requested, please check values.")
        pass
        return True
    elif not current[0].isdigit():  # current is package name rather than version string, i.e. current='visaionlibrary.utils.yoloe'
        try:
            name = current  # assigned package name to 'name' arg
            current = metadata.version(current)  # get version string from package name
        except metadata.PackageNotFoundError as e:
            if hard:
                raise ModuleNotFoundError(emojis(f"WARNING ⚠️ {current} package is required but not installed")) from e
            else:
                return False

    if not required:  # if required is '' or None
        return True

    if "sys_platform" in required and (  # i.e. required='<2.4.0,>=1.8.0; sys_platform == "win32"'
        (WINDOWS and "win32" not in required)
        or (LINUX and "linux" not in required)
        or (MACOS and "macos" not in required and "darwin" not in required)
    ):
        return True

    op = ""
    version = ""
    result = True
    c = parse_version(current)  # '1.2.3' -> (1, 2, 3)
    for r in required.strip(",").split(","):
        op, version = re.match(r"([^0-9]*)([\d.]+)", r).groups()  # split '>=22.04' -> ('>=', '22.04')
        if not op:
            op = ">="  # assume >= if no op passed
        v = parse_version(version)  # '1.2.3' -> (1, 2, 3)
        if op == "==" and c != v:
            result = False
        elif op == "!=" and c == v:
            result = False
        elif op == ">=" and not (c >= v):
            result = False
        elif op == "<=" and not (c <= v):
            result = False
        elif op == ">" and not (c > v):
            result = False
        elif op == "<" and not (c < v):
            result = False
    if not result:
        warning = f"WARNING ⚠️ {name}{op}{version} is required, but {name}=={current} is currently installed {msg}"
        if hard:
            raise ModuleNotFoundError(emojis(warning))  # assert version requirements met
        if verbose:
            # LOGGER.warning(warning)
            pass
    return result



def check_python(minimum: str = "3.8.0", hard: bool = True, verbose: bool = False) -> bool:
    """
    Check current python version against the required minimum version.

    Args:
        minimum (str): Required minimum version of python.
        hard (bool, optional): If True, raise an AssertionError if the requirement is not met.
        verbose (bool, optional): If True, print warning message if requirement is not met.

    Returns:
        (bool): Whether the installed Python version meets the minimum constraints.
    """
    return check_version(PYTHON_VERSION, minimum, name="Python", hard=hard, verbose=verbose)


def check_suffix(file="yolo11n.pt", suffix=".pt", msg=""):
    """Check file(s) for acceptable suffix."""
    if file and suffix:
        if isinstance(suffix, str):
            suffix = (suffix,)
        for f in file if isinstance(file, (list, tuple)) else [file]:
            s = Path(f).suffix.lower().strip()  # file suffix
            if len(s):
                assert s in suffix, f"{msg}{f} acceptable suffix is {suffix}, not {s}"


def check_yolov5u_filename(file: str, verbose: bool = True):
    """Replace legacy YOLOv5 filenames with updated YOLOv5u filenames."""
    if "yolov3" in file or "yolov5" in file:
        if "u.yaml" in file:
            file = file.replace("u.yaml", ".yaml")  # i.e. yolov5nu.yaml -> yolov5n.yaml
        elif ".pt" in file and "u" not in file:
            original_file = file
            file = re.sub(r"(.*yolov5([nsmlx]))\.pt", "\\1u.pt", file)  # i.e. yolov5n.pt -> yolov5nu.pt
            file = re.sub(r"(.*yolov5([nsmlx])6)\.pt", "\\1u.pt", file)  # i.e. yolov5n6.pt -> yolov5n6u.pt
            file = re.sub(r"(.*yolov3(|-tiny|-spp))\.pt", "\\1u.pt", file)  # i.e. yolov3-spp.pt -> yolov3-sppu.pt
            # if file != original_file and verbose:
            #     LOGGER.info(
            #         f"PRO TIP 💡 Replace 'model={original_file}' with new 'model={file}'.\nYOLOv5 'u' models are "
            #         f"trained with https://github.com/visaionlibrary.utils.yoloe/visaionlibrary.utils.yoloe and feature improved performance vs "
            #         f"standard YOLOv5 models trained with https://github.com/visaionlibrary.utils.yoloe/yolov5.\n"
            #     )
    return file


def check_model_file_from_stem(model="yolov8n"):
    """Return a model filename from a valid model stem."""
    if model and not Path(model).suffix and Path(model).stem in downloads.GITHUB_ASSETS_STEMS:
        return Path(model).with_suffix(".pt")  # add suffix, i.e. yolov8n -> yolov8n.pt
    else:
        return model


def check_file(file, suffix="", download=True, download_dir=".", hard=True):
    """Search/download file (if necessary) and return path."""
    check_suffix(file, suffix)  # optional
    file = str(file).strip()  # convert to string and strip spaces
    file = check_yolov5u_filename(file)  # yolov5n -> yolov5nu
    if (
        not file
        or ("://" not in file and Path(file).exists())  # '://' check required in Windows Python<3.10
        or file.lower().startswith("grpc://")
    ):  # file exists or gRPC Triton images
        return file
    elif download and file.lower().startswith(("https://", "http://", "rtsp://", "rtmp://", "tcp://")):  # download
        url = file  # warning: Pathlib turns :// -> :/
        file = Path(download_dir) / url2file(file)  # '%2F' to '/', split https://url.com/file.txt?auth
        if file.exists():
            # LOGGER.info(f"Found {clean_url(url)} locally at {file}")  # file already exists
            pass
        else:
            downloads.safe_download(url=url, file=file, unzip=False)
        return str(file)
    else:  # search
        files = glob.glob(str(ROOT / "**" / file), recursive=True) or glob.glob(str(ROOT.parent / file))  # find file
        if not files and hard:
            raise FileNotFoundError(f"'{file}' does not exist")
        elif len(files) > 1 and hard:
            raise FileNotFoundError(f"Multiple files match '{file}', specify exact path: {files}")
        return files[0] if len(files) else []  # return file


def check_yaml(file, suffix=(".yaml", ".yml"), hard=True):
    """Search/download YAML file (if necessary) and return path, checking suffix."""
    return check_file(file, suffix, hard=hard)


def check_is_path_safe(basedir, path):
    """
    Check if the resolved path is under the intended directory to prevent path traversal.

    Args:
        basedir (Path | str): The intended directory.
        path (Path | str): The path to check.

    Returns:
        (bool): True if the path is safe, False otherwise.
    """
    base_dir_resolved = Path(basedir).resolve()
    path_resolved = Path(path).resolve()

    return path_resolved.exists() and path_resolved.parts[: len(base_dir_resolved.parts)] == base_dir_resolved.parts


def check_yolo(verbose=True, device=""):
    """Return a human-readable YOLO software and hardware summary."""
    import psutil

    if IS_COLAB:
        shutil.rmtree("sample_data", ignore_errors=True)  # remove colab /sample_data directory

    if verbose:
        # System info
        gib = 1 << 30  # bytes per GiB
        ram = psutil.virtual_memory().total
        total, used, free = shutil.disk_usage("/")
        s = f"({os.cpu_count()} CPUs, {ram / gib:.1f} GB RAM, {(total - free) / gib:.1f}/{total / gib:.1f} GB disk)"
        try:
            from IPython import display

            display.clear_output()  # clear display if notebook
        except ImportError:
            pass
    else:
        s = ""

    torch.device(device)
    # LOGGER.info(f"Setup complete ✅ {s}")
